/******************************************************************************
 * Copyright (C) 2003 David Tam, Catalin Drula, Winnie Tsang
 * tamda@eecg.toronto.edu
 * catalin@cs.toronto.edu
 * wtsang@dgp.toronto.edu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *****************************************************************************/
#ifndef DVS_H
#define DVS_H

#include <linux/sched.h> /* Needed for task_t */


/******************************************************************************
 * VERY IMPORTANT VERY IMPORTANT VERY IMPORTANT VERY IMPORTANT VERY IMPORTANT
 *
 * This macro should only be defined if you plan to use the embedded version of
 * DVS instead of the loadable kernel module version.
 *
 * You ***MUST NOT*** define this macro if you plan to use the loadable kernel
 * module version of DVS.  Otherwise, you will have conflicts.
 */
/*#define DVS_EMBEDDED*/
/*
 * VERY IMPORTANT VERY IMPORTANT VERY IMPORTANT VERY IMPORTANT VERY IMPORTANT
 *****************************************************************************/


/*
 * How often DVS algorithm is run, e.g. every 1000 jiffies == 1000 ms.
 * This is the max period between invocation of the DVS algorithm.
 */
#define DVSFREQ 1000

/*
 *Size of history to keep. Try to use powers of 2 for low-cost "modding"
 * This value is critically related to DVSFREQ. There should be enough
 * buckets to store utilization history in between calls to DVS algorithm.
 * For DVSFREQ == 1000, I only need 200 buckets to allow for storage of
 * 2 DVS alg intervals. I need 2 DVS alg intervals to support a DVS window
 * size of 2 theoretical DVS intervals. Need only 200 buckets because the
 * min bucket size is 10 ms (=10 jiffies)
 */
#define DVSBUCKETS 256

/*
 * Holds quantum size given to this task and amount of unused quantum upon task
 * switch.
 */
typedef struct dvs_quantum_usage {
	unsigned int orig_time_slice; /* In units of jiffies */
	unsigned long long ns_left;        /* In units of nanoseconds */
} dvs_quantum_usage_t;

/*
 * Here is a small static array for holding the quantum usage history.
 * This will be used as a circular buffer.
 */
extern dvs_quantum_usage_t dvs_qu_array[DVSBUCKETS];

/* The "current pointer" of the above circular buffer */
extern int dvs_qu_index;

/* current speed of the cpu, defined in dvs.c */
extern int dvs_speed;

/* Function that records the load of the task */
/* Get Called by dvs_record_usage() */
inline extern void task_load_record(task_t *p, unsigned long long ns_left, int speed){
        
        p->task_load_array[p->task_load_index].orig_time_slice = p->orig_time_slice;
        p->task_load_array[p->task_load_index].ns_left = ns_left;
        p->task_load_array[p->task_load_index].speed = speed;
        p->task_load_index = (p->task_load_index+1) % TASKLOADSIZE;

#if 0
        /* Debugging stuff delete this junk */
        printk(KERN_ERR " time_slice=%u, ns_left=%llu, speed=%d",p->orig_time_slice, ns_left, speed);
#endif /* 0 */

}

/*
 * Function that records quantum usage.
 * I assume that this function won't be run concurrently because I don't use
 * locks and I am modifying data.
 */
inline extern void dvs_record_usage(task_t *p, unsigned long long ns_left) {
	dvs_qu_array[dvs_qu_index].orig_time_slice = p->orig_time_slice;
	dvs_qu_array[dvs_qu_index].ns_left = ns_left;
	dvs_qu_index = (dvs_qu_index + 1) % DVSBUCKETS;
        /* also record into task_load_array */
        task_load_record(p, ns_left, dvs_speed);  

}


/******************************************************************************
 * This section is included only for the embedded version of DVS.
 */
#ifdef DVS_EMBEDDED

/* When == 0, then the DVS algorithm is run */
extern int dvs_wait;

/* Runs the appropriate CPU governor that is embedded in kernel.
 * This means that the DVS loadable kernel module is not being used. */
void dvs_runalg(task_t *p);

#endif /* DVS_EMBEDDED */
/*
 * This section is included only for the embedded version of DVS.
 *****************************************************************************/


#endif /* DVS_H */
