/******************************************************************************
 * Copyright (C) 2003 David Tam, Catalin Drula, Winnie Tsang
 * tamda@eecg.toronto.edu
 * catalin@cs.toronto.edu
 * wtsang@dgp.toronto.edu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *****************************************************************************/
#include <linux/module.h>
#include <linux/init.h>
#include <linux/cpufreq.h>
#include <linux/workqueue.h>
#include <asm/msr.h>
#include <linux/kernel_stat.h>
#include <linux/sysfs.h>
#include <linux/string.h>

#include <linux/dvs.h> /* For accessing kernel dvs structs */

#define DEBUG

#define DEFAULT_ALGORITHM 0 /* number of default algorithm */
#define ALG_NAME_LENGTH 20  /* max length of algorithm name */
#define NR_ALG 4            /* number of algorithms */

#define DEFAULT_INTERVAL_LENGTH 50    /* in jiffies */
#define NS_TO_JIFFIES(TIME)      ((TIME) / (1000000000 / HZ))

int interval_length_jiffies;
struct cpufreq_policy policies[NR_CPUS];

/* The decision returned by the DVS algorithm */
struct cpufreq_dvs_decision {
  // INCREASE means increase the speed by a number of steps
  // DECREASE is similar
  // SETSPEED means set a certain speed
  enum type {INCREASE, DECREASE, SETSPEED} type; 
  int steps;  // for INCREASE and DECREASE
  int speed;  // for SETSPEED
};

int current_alg;      /* number of current algorithm */

/* forward declarations */
struct cpufreq_dvs_decision *avg(int cpu);
void avg_init(int cpu);
void avg_cleanup(int cpu);

/* Past algorithm */
struct cpufreq_dvs_decision *past(int cpu);
void past_init(int cpu);
void past_cleanup(int cpu);

/* Future algorithm */
struct cpufreq_dvs_decision *future(int cpu);
void future_init(int cpu);
void future_cleanup(int cpu);

/* Peak algorithm */
struct cpufreq_dvs_decision *peak(int cpu);
void peak_init(int cpu);
void peak_cleanup(int cpu);

/* Functions to process kernel dvs data */
int get_idlecycles(int interval);
int get_runcycles(int interval);
int get_excesscycles(int interval);
int get_idlecycles_future(void);

/* Different version of function to get utilization */
unsigned int get_current_utilization_real(int cpu);
unsigned int get_current_utilization_hype(int cpu);
unsigned int get_current_utilization_futurehype(int cpu);
/* Function pointer that points to 1 of the 3 above functions */
unsigned int (*fp_get_cur_util)(int cpu);

/* this hold information for an algorithm */
struct dvs_algorithm {
  struct cpufreq_dvs_decision * (*algorithm) (int);
  void (*init)(int);
  void (*cleanup)(int);
  char name[ALG_NAME_LENGTH];
} algorithms[NR_ALG] = { { .algorithm = avg, .init = avg_init,
			   .cleanup = avg_cleanup, .name = "avg" } ,
			 { .algorithm = past, .init = past_init,
			   .cleanup = past_cleanup, .name = "past" } ,
			 { .algorithm = future, .init = future_init,
			   .cleanup = future_cleanup, .name = "future" } ,
			 { .algorithm = peak, .init = peak_init,
			   .cleanup = peak_cleanup, .name = "peak" } };



/************************* DVS algorithms ***************************/
/***** This section contains the DVS algorithms *********************/
/********************************************************************/

/************* sysfs variables and definitions *******************/

static ssize_t show_avg_n(struct cpufreq_policy *, char *);
static ssize_t store_avg_n(struct cpufreq_policy *, const char *, size_t);
static struct freq_attr freq_attr_avg_n = {
  .attr = { .name = "avg_n", .mode = 0644 },
  .show = show_avg_n,
  .store = store_avg_n,
};

static ssize_t show_hyst_min(struct cpufreq_policy *, char *);
static ssize_t store_hyst_min(struct cpufreq_policy *, const char *, size_t);
static struct freq_attr freq_attr_hyst_min = {
  .attr = { .name = "hystmin", .mode = 0644 },
  .show = show_hyst_min,
  .store = store_hyst_min,
};

static ssize_t show_hyst_max(struct cpufreq_policy *, char *);
static ssize_t store_hyst_max(struct cpufreq_policy *, const char *, size_t);
static struct freq_attr freq_attr_hyst_max = {
  .attr = { .name = "hystmax", .mode = 0644 },
  .show = show_hyst_max,
  .store = store_hyst_max,
};

static ssize_t show_past_incr(struct cpufreq_policy *, char *);
static ssize_t store_past_incr(struct cpufreq_policy *, const char *, size_t);
static struct freq_attr freq_attr_past_incr = {
  .attr = { .name = "past_incr", .mode = 0644 },
  .show = show_past_incr,
  .store = store_past_incr,
};

static ssize_t show_past_basesub(struct cpufreq_policy *, char *);
static ssize_t store_past_basesub(struct cpufreq_policy *, const char *, size_t);
static struct freq_attr freq_attr_past_basesub = {
  .attr = { .name = "past_basesub", .mode = 0644 },
  .show = show_past_basesub,
  .store = store_past_basesub,
};

/* Allows you to select 1 of 3 implementations to get utilization info */
static ssize_t show_getcurutil(struct cpufreq_policy *, char *);
static ssize_t store_getcurutil(struct cpufreq_policy *, const char *, size_t);
static struct freq_attr freq_attr_getcurutil = {
  .attr = { .name = "getcurutil", .mode = 0644 },
  .show = show_getcurutil,
  .store = store_getcurutil,
};


/********Structures & variables used by the past,peak,future algs*************/

/*** default values for "avg" parameters */
/*** default values HYST MIN & MAX for past & peak algs too */
#define DEFAULT_AVG_N 5
#define DEFAULT_HYST_MIN 50
#define DEFAULT_HYST_MAX 70

/* hyst_min and hyst_max used by past alg too */
int avg_n, hyst_min, hyst_max;
int prev_forecast[NR_CPUS];
/* "decision" variable is used by past and peak algs too */
/* Warning: decision should be an array of size NR_CPUS? */
struct cpufreq_dvs_decision decision;



/**************** AVG_n algorithm **********************************/

/* initialization for avg algorithm */
void avg_init(int cpu) {
    /* set defaults for the variables */
    /* note: this should be cpu-specific, but for the moment this
     * is fine */
    avg_n = DEFAULT_AVG_N;
    hyst_min = DEFAULT_HYST_MIN;
    hyst_max = DEFAULT_HYST_MAX;

    prev_forecast[cpu]=100;

    /* Set function pointer */
    fp_get_cur_util = get_current_utilization_real;

    /* create sysfs files */
    sysfs_create_file(&policies[cpu].kobj, &freq_attr_hyst_max.attr);
    sysfs_create_file(&policies[cpu].kobj, &freq_attr_hyst_min.attr);
    sysfs_create_file(&policies[cpu].kobj, &freq_attr_avg_n.attr);
    sysfs_create_file(&policies[cpu].kobj, &freq_attr_getcurutil.attr);
}

/* cleanup function for avg algorithm */
void avg_cleanup(int cpu) {
    sysfs_remove_file (&policies[cpu].kobj, &freq_attr_hyst_max.attr);
    sysfs_remove_file (&policies[cpu].kobj, &freq_attr_hyst_min.attr);
    sysfs_remove_file (&policies[cpu].kobj, &freq_attr_avg_n.attr);
    sysfs_remove_file (&policies[cpu].kobj, &freq_attr_getcurutil.attr);
}

/* main function implementing the AVG_n algorithm */
struct cpufreq_dvs_decision *avg(int cpu) {

  /* this is the AVG_n algorithm as described in Grunwald et al. */
  if (cpu >= 0 && cpu < NR_CPUS) {
    unsigned int util = (*fp_get_cur_util)(cpu);
    int forecast = (avg_n * prev_forecast[cpu] + util) / (avg_n+1);
    prev_forecast[cpu] = forecast;

    if (forecast < hyst_min) {
      decision.type = DECREASE;
      decision.steps = 1;
      return (&decision);
    } else if (forecast > hyst_max) {
      decision.type = INCREASE;
      decision.steps = 1;
      return (&decision);
    } else return NULL;
  } else return NULL;
}


/********Structures & variables used by the past,peak,future algs*************/

#define CONSTANT 1    /* CONSTANT for Peak Algorithm */
#define MIN_SPEED 10	 /* minimum percentage of speed */
#define PAST_INCR 20   /* Size of increment in past/future algorithm */
#define PAST_BASESUB 60 /* Size of base number used in decrement operation in
			 * past/future alg */

/***********************************************************************
 * structure for store interval statistics for past,peak,future algs
 *********************************************************************/
struct dvs_struct{
	int run_percent;
	int speed;
};

/* stores the current */
struct dvs_struct dvs_current[NR_CPUS];

/* Holds current size of increment and base subtraction parameters for
 * past/future algorithm */
int past_incr;
int past_basesub;

/* Function prototype */
struct cpufreq_dvs_decision *futurepast(int cpu, int interval);


/**************** past algorithm **********************************/

/* main function implementing the past algorithm */
/* PAST */
/* Compute the speed and predicted run_percent and stores these values
 * into *dvs_current struct and return point to decision too.
 */
struct cpufreq_dvs_decision *past(int cpu) {
	return futurepast(cpu, 0);
}

/* initialization for past algorithm */
void past_init(int cpu) {
	dvs_current[cpu].speed = 100;
	interval_length_jiffies = DVSFREQ;
	hyst_min = DEFAULT_HYST_MIN;
	hyst_max = DEFAULT_HYST_MAX;
	past_incr = PAST_INCR;
	past_basesub = PAST_BASESUB;
	
	/* create sysfs files */
	sysfs_create_file(&policies[cpu].kobj, &freq_attr_hyst_max.attr);
	sysfs_create_file(&policies[cpu].kobj, &freq_attr_hyst_min.attr);
	sysfs_create_file(&policies[cpu].kobj, &freq_attr_past_incr.attr);
	sysfs_create_file(&policies[cpu].kobj, &freq_attr_past_basesub.attr);
}

/* cleanup function for past algorithm */
void past_cleanup(int cpu) {	
	sysfs_remove_file (&policies[cpu].kobj, &freq_attr_hyst_max.attr);
	sysfs_remove_file (&policies[cpu].kobj, &freq_attr_hyst_min.attr);
	sysfs_remove_file (&policies[cpu].kobj, &freq_attr_past_incr.attr);
	sysfs_remove_file (&policies[cpu].kobj, &freq_attr_past_basesub.attr);
}

/**************** future algorithm **********************************/

/* main function implementing the future algorithm */
/* PAST */
/* Compute the speed and predicted run_percent and stores these values
 * into *dvs_current struct and return point to decision too.
 */
struct cpufreq_dvs_decision *future(int cpu) {
	return futurepast(cpu, -1);
}

/* initialization for future algorithm */
void future_init(int cpu) {
	past_init(cpu);
}

/* cleanup function for future algorithm */
void future_cleanup(int cpu) {	
	past_cleanup(cpu);
}


/***** Main support function for past and future algorithms ******************/

/* Main support function implementing the past and future algorithms */
/* PAST/FUTURE */
/* Compute the speed and predicted run_percent and stores these values
 * into *dvs_current struct and return point to decision too.
 * Parameters:
 * interval == -1 use pseudo-future algorithm
 * interval == 0  use past algorithm
 */
struct cpufreq_dvs_decision *futurepast(int cpu, int interval) {
	/* Declare local variables */
	int run_cycle;
	int idle_cycle;
	int excess_cycle;
	int excess_percent;
	int next_excess_cycle;

	/* Check parameter boundary */
	if (interval < -1 || interval > 0) {
		return &decision;
	}

	/* Get Pt run percent */
	run_cycle=get_runcycles(interval);
	idle_cycle=get_idlecycles(interval);
	excess_cycle=get_excesscycles(interval);

	dvs_current[cpu].run_percent = (run_cycle + excess_cycle)*100/
		(run_cycle + idle_cycle);
	
#ifdef FULLDEBUG
	printk(KERN_ERR "run_cycle = %d, idle_cycle = %d, excess_cycle = %d\n",
	       run_cycle, idle_cycle, excess_cycle);
#endif /* FULLDEBUG */

	/* Compute excess cycles */
	next_excess_cycle = ((excess_cycle+run_cycle)*100-
			     dvs_current[cpu].speed*(run_cycle+idle_cycle))/
	                     100;
	if (next_excess_cycle < 0) {
		next_excess_cycle = 0;
	}
	excess_percent = next_excess_cycle*100/(run_cycle + idle_cycle);

	/* Set Speed */
	if (excess_cycle > idle_cycle) {
		dvs_current[cpu].speed = 100;
	}
	else if (dvs_current[cpu].run_percent > hyst_max) {
		dvs_current[cpu].speed = dvs_current[cpu].speed + past_incr;
	}
	else if (dvs_current[cpu].run_percent < hyst_min) {
		dvs_current[cpu].speed = dvs_current[cpu].speed -
			(past_basesub - dvs_current[cpu].run_percent);
	}
	if (dvs_current[cpu].speed > 100) {
		dvs_current[cpu].speed = 100;
	}
	if (dvs_current[cpu].speed < MIN_SPEED) {
		dvs_current[cpu].speed = MIN_SPEED;
	}

	decision.type = SETSPEED;
	decision.speed = dvs_current[cpu].speed;
	return &decision;
}


/**************** peak algorithm **********************************/

/* Main function implementing the peak algorithm */
/* PEAK */
/* Compute the speed and predicted run_percent and stores these values
   into *dvs_current[cpu] struct */
struct cpufreq_dvs_decision * peak(int cpu) {

	/* Declare local variables */
	int run_cycle;
	int idle_cycle;
	int excess_cycle;
	int excess_percent;
	int next_excess_cycle;
	int run_percent[2];

	/* Get Pt-1 run percent */
	run_cycle=get_runcycles(1);
	idle_cycle=get_idlecycles(1);
	excess_cycle=get_excesscycles(1);
#ifdef FULLDEBUG
	printk(KERN_ERR "run_cyc1 = %d, idle_cyc1 = %d, excess_cyc1 = %d\n",
	       run_cycle, idle_cycle, excess_cycle);
#endif /* FULLDEBUG */
	run_percent[1] = (run_cycle + excess_cycle)*100/(run_cycle +
							 idle_cycle);
	
	/* Get Pt run percent */
	run_cycle=get_runcycles(0);
	idle_cycle=get_idlecycles(0);
	excess_cycle=get_excesscycles(0);
#ifdef FULLDEBUG 
	printk(KERN_ERR "run_cyc0 = %d, idle_cyc0 = %d, excess_cyc0 = %d\n",
	       run_cycle, idle_cycle, excess_cycle);
#endif /* FULLDEBUG */
	run_percent[0] = (run_cycle + excess_cycle)*100/(run_cycle +
							 idle_cycle);

	/* Compute excess cycle outgoing */
	next_excess_cycle = ((excess_cycle+run_cycle)*100-
			     dvs_current[cpu].speed*(run_cycle+idle_cycle))/
		             100;
	if (next_excess_cycle < 0) {
		next_excess_cycle = 0;
	}
	excess_percent = next_excess_cycle*100/(run_cycle + idle_cycle);

	/* Prediction, set dvs_current[cpu]'s run percent based on past two
	 * intervals statistics*/
	/* Pt > Pt-1 */
	if (run_percent[0] > run_percent[1]) {
		if (run_percent[1] > 10) {
			dvs_current[cpu].run_percent = run_percent[1];
		}
		else {
			dvs_current[cpu].run_percent = 10;
		}
	}
	/* Pt < Pt-1 */
	else if (run_percent[0] < run_percent[1]) {
		if (run_percent[1] > 10) {
			dvs_current[cpu].run_percent = run_percent[0];
		}
		else {
			dvs_current[cpu].run_percent = 10;
		}
	}
	/* Pt = Pt-1 */
	else {
		if (run_percent[0]==100) {
			dvs_current[cpu].run_percent = 40;
		}
		else{
			dvs_current[cpu].run_percent=run_percent[0];
		}
	}
	
	/* Set Speed */
	dvs_current[cpu].speed = dvs_current[cpu].run_percent +
		CONSTANT*excess_percent;
	if (dvs_current[cpu].speed > 100) {
		dvs_current[cpu].speed = 100;
	}
	if (dvs_current[cpu].speed < MIN_SPEED) {
		dvs_current[cpu].speed = MIN_SPEED;
	}
	
	decision.type = SETSPEED;
	decision.speed = dvs_current[cpu].speed;
	return &decision;
}

/* initialization for peak algorithm */
void peak_init(int cpu) {
	dvs_current[cpu].speed = 100;
	interval_length_jiffies = DVSFREQ;
}

/* cleanup function for peak algorithm */
void peak_cleanup(int cpu) {
	/* Nothing to do here */
}


/**************************** DVS ***************************************/
/****** This section contains the "algorithm-independent" functions *****/
/************************************************************************/

/* Returns the current utilization of the cpu as an integer
 * between 0 and 100 
 */
unsigned int get_current_utilization_real(int cpu) {
  static struct cpu_usage_stat oldstat;
  struct cpu_usage_stat newstat;
  unsigned int util;

  newstat = kstat_cpu(cpu).cpustat;
  util = (newstat.user - oldstat.user) + (newstat.system - oldstat.system) +
    (newstat.irq - oldstat.irq) + (newstat.softirq - oldstat.softirq) + 
    (newstat.nice - oldstat.nice) + (newstat.iowait - oldstat.iowait);
  util = (100 * util) / (util + newstat.idle - oldstat.idle);
  
  oldstat = newstat;
  return util;
}


/* Returns the current theoretical utilization of the cpu as an integer
 * between 0 and 100. Currently works for uniprocessor only.
 * This version uses the kernel dvs data to generate the result.
 */
unsigned int get_current_utilization_hype(int cpu) {
  return (100 * get_runcycles(0)) / interval_length_jiffies;
}

/* Returns the current pseudo-future theoretical utilization of the cpu as an
 * integer between 0 and 100. Currently works for uniprocessor only.
 * This version uses the kernel dvs data to generate the result.
 */
unsigned int get_current_utilization_futurehype(int cpu) {
  return (100 * get_runcycles(-1)) / interval_length_jiffies;
}

/* this is a table of available frequencies; this needs to be changed to
 * a more general mechanism */
#define MAX_NR_FREQ 20
int avail_freq[MAX_NR_FREQ], nfreq, curfreq;

/* scale: puts "into practice" the decision taken by the DVS algorithm */
void scale(struct cpufreq_dvs_decision *dvs_decision, int cpu) {
  unsigned int speed;

  /* if the decision was not to change anything, return */
  if (dvs_decision == NULL) return;

  /* set an absolute frequency */
  if (dvs_decision->type == SETSPEED) {
    unsigned int perc_speed = dvs_decision->speed;
    //    speed = (((policies[cpu].max - policies[cpu].min) * perc_speed)/100)
    //  + policies[cpu].min;
    speed = (policies[cpu].max  * perc_speed)/100;
    if (speed <= 1200000)
      cpufreq_driver_target(&policies[cpu], speed, CPUFREQ_RELATION_H);
    else 
      cpufreq_driver_target(&policies[cpu], 1600000, CPUFREQ_RELATION_H);

#ifdef FULLDEBUG
    printk(KERN_DEBUG
	   "cpufreq_dvs: changed speed to %d\n", speed);
#endif
  }
  
  /* decrease speed by one step */
  if (dvs_decision->type == DECREASE && curfreq > 0) {
    curfreq--;
    cpufreq_driver_target(&policies[cpu], avail_freq[curfreq], 
			  CPUFREQ_RELATION_H);
#ifdef DEBUG
    printk(KERN_DEBUG 
	   "cpufreq_dvs: decreased speed to %d\n", avail_freq[curfreq]);
#endif
  }
  
  /* increase speed by one step */
  if (dvs_decision->type == INCREASE && curfreq < nfreq-1) {
      curfreq++;
      cpufreq_driver_target(&policies[cpu], avail_freq[curfreq], 
			    CPUFREQ_RELATION_H);
#ifdef DEBUG
      printk(KERN_DEBUG "cpufreq_dvs: increased speed to %d\n", 
	     avail_freq[curfreq]);
#endif
  }

  /* Updating our DVS kernel variable and our module variable */
  dvs_speed = 100.0 * policies[cpu].cur / policies[cpu].max;
  dvs_current[cpu].speed = dvs_speed;
}

/* work struct */
void dvs_work(void *p);
DECLARE_WORK(cpufreq_dvs_work, dvs_work, NULL);

/* 
 * dvs_work is called every interval_length_jiffies 
 * It runs the dvs algorithm
 */
void dvs_work(void *p) {
  int cpu;

  /* scale speed for each cpu if needed */
  for (cpu=0; cpu<NR_CPUS; cpu++)
    scale(algorithms[current_alg].algorithm(cpu), cpu);

  /* reschedule myself to run later */
  schedule_delayed_work(&cpufreq_dvs_work, interval_length_jiffies);
}

/*************************** governor ************************************/
/********** This section contains the governor functions *****************/
/*************************************************************************/

/*** sysfs variables and definitions ***/

static ssize_t show_interval(struct cpufreq_policy *, char *);
static ssize_t store_interval (struct cpufreq_policy *, const char *, size_t);
static struct freq_attr freq_attr_interval = {
  .attr = { .name = "interval", .mode = 0644 },
  .show = show_interval,
  .store = store_interval,
};

static ssize_t show_available_algorithms(struct cpufreq_policy *, char *);
static struct freq_attr freq_attr_available_algorithms = {
  .attr = { .name = "available_algorithms", .mode = 0444 },
  .show = show_available_algorithms,
};

static ssize_t show_dvs_algorithm(struct cpufreq_policy *, char *);
static ssize_t store_dvs_algorithm(struct cpufreq_policy *, const char *, size_t);
static struct freq_attr freq_attr_dvs_algorithm = {
  .attr = { .name = "dvs_algorithm", .mode = 0644 },
  .show = show_dvs_algorithm,
  .store = store_dvs_algorithm,
};

/*** The governor function (event handler) ***/
static int cpufreq_governor_dvs(struct cpufreq_policy *policy,
				unsigned int event) {
  int cpu = policy->cpu;
  unsigned int latency_jiffies;

  switch (event) {

    // Governor is started
  case CPUFREQ_GOV_START:
    #ifdef DEBUG
    printk(KERN_INFO "cpufreq_dvs: CPUFREQ_GOV_START\n");
    #endif
    
    /* Store policy for this CPU */
    policies[cpu] = *policy;

    // initialize DVS data structure
    algorithms[current_alg].init(cpu);
    
    // Handle the case where transition latency is greater than
    // our desired interval length
    latency_jiffies =
      NS_TO_JIFFIES(policy->cpuinfo.transition_latency);
    // David Tam Nov 24, 2003.
    // Original:
    //if (latency_jiffies > 0 && latency_jiffies < interval_length_jiffies)
    // New:
    /*    if (latency_jiffies != CPUFREQ_ETERNAL && 
	latency_jiffies > 0 && 
	latency_jiffies > interval_length_jiffies)
	interval_length_jiffies = latency_jiffies;*/

    #ifdef DEBUG
    printk(KERN_INFO "cpufreq_dvs: Interval length is %u jiffies\n", 
	   interval_length_jiffies);
    #endif

    /* create sysfs files for general governor variables */
    sysfs_create_file(&policy->kobj, &freq_attr_interval.attr);
    sysfs_create_file(&policy->kobj, &freq_attr_dvs_algorithm.attr);
    sysfs_create_file(&policy->kobj, &freq_attr_available_algorithms.attr);

    /* Schedule DVS algorithm */
    schedule_delayed_work(&cpufreq_dvs_work, interval_length_jiffies);
    break;

    // Governor is stopped
  case CPUFREQ_GOV_STOP:
    #ifdef DEBUG
    printk(KERN_INFO "cpufreq_dvs: CPUFREQ_GOV_STOP\n");
    #endif

    /* Cancel scheduled run of DVS algorithm */
    cancel_delayed_work(&cpufreq_dvs_work);

    /* remove sysfs files */
    sysfs_remove_file (&policy->kobj, &freq_attr_interval.attr);
    sysfs_remove_file (&policy->kobj, &freq_attr_dvs_algorithm.attr);
    sysfs_remove_file (&policy->kobj, &freq_attr_available_algorithms.attr);

    /* call DVS algorithm cleanup */
    algorithms[current_alg].cleanup(cpu);
    break;

    // Governor is notified that the frequency limits changed
  case CPUFREQ_GOV_LIMITS:
    #ifdef DEBUG
    printk(KERN_INFO "cpufreq_dvs: CPUFREQ_GOV_LIMITS\n");
    #endif

    /* Update policy for this cpu */
    policies[cpu]=*policy;
    break;
  }
  return 0;
}

/* Declare the governor structure */
struct cpufreq_governor cpufreq_gov_dvs = {
  .name = "dvs",
  .governor = cpufreq_governor_dvs,
  .owner = THIS_MODULE,
};

/************************************* module *****************************************/
/* This section contains module initialization and cleanup functions */
/**************************************************************************************/

/* Initialization function for the module */
int cpufreq_dvs_init(void) {
#ifdef DEBUG
  printk(KERN_INFO "cpufreq_dvs: loading\n");
#endif

  //this is a table of frequencies to be used with the p4-clockmod driver 
  //running on my P4-M 1.6GHz 
  //Note: this should be changed to an architecture/driver independent mechanism
  //  nfreq = 8;
  //curfreq = nfreq-1;
  //avail_freq[0] = 199884;avail_freq[1] = 399768;avail_freq[2] =599653;avail_freq[3] =799537;
  //avail_freq[4] = 999421;avail_freq[5] = 1199306;avail_freq[6]=1399190; avail_freq[7]=1599075;
  //table of frequencies for P4-M 1.6 GHZ with the speedstep-ich driver
  nfreq = 2;
  curfreq = nfreq-1;
  avail_freq[0] = 1150000; avail_freq[1]=1600000;

  //set the default algorithm
  current_alg=DEFAULT_ALGORITHM;

  //set the default interval length
  interval_length_jiffies = DEFAULT_INTERVAL_LENGTH;

  //register using cpufreq_register_governor
  return cpufreq_register_governor(&cpufreq_gov_dvs);
}

/* Cleanup function for the module */
static void __exit cpufreq_dvs_exit(void) {
  // unregister governor
  cpufreq_unregister_governor(&cpufreq_gov_dvs);

#ifdef DEBUG
  printk(KERN_INFO "cpufreq_dvs unloaded\n");
#endif
}

/************************************* sysfs   ****************************************/
/* This section contains sysfs functions */
/**************************************************************************************/

/************ General governor variables ******************/

/**** interval ****/
/* This is for getting/setting the length of the interval at which we run the algorithm */
static ssize_t show_interval(struct cpufreq_policy *policy, char *buf)
{
  return sprintf (buf, "%d\n", interval_length_jiffies);
}
                                                                                                           
static ssize_t
store_interval (struct cpufreq_policy *policy, const char *buf, size_t count)
{
  unsigned int ret, latency_jiffies;
 
  ret = sscanf (buf, "%u", &interval_length_jiffies);
  if (ret != 1)
    return -EINVAL;

  latency_jiffies =
    NS_TO_JIFFIES(policy->cpuinfo.transition_latency);
  if (latency_jiffies > 0 && latency_jiffies < interval_length_jiffies)
    interval_length_jiffies = latency_jiffies;

  return count;
}
                                                                                                                          
/*** available_algorithms ***/
/* Provides a list of available dvs algorithms */
static ssize_t show_available_algorithms(struct cpufreq_policy *policy, char *buf)
{
  char *aux = buf;
  int count=0, i;
  for (i=0; i<NR_ALG; i++) {
    count += sprintf(aux, "%s\n", algorithms[i].name);    
    aux += count;
  }

  return count;
}
                                                                                                           
/*** dvs_algorithm ***/
/* Get/set for the current algorithm */
static ssize_t show_dvs_algorithm(struct cpufreq_policy *policy, char *buf)
{
  return sprintf (buf, "%s\n", algorithms[current_alg].name);
}

static ssize_t
store_dvs_algorithm(struct cpufreq_policy *policy, const char *buf, size_t count)
{
  int i=0;
  char s[80];

  sscanf(buf, "%s", s);
  while (i < NR_ALG) {
    if (!strcmp(s, algorithms[i].name)) {
      int j;
      for (j=0; j<NR_CPUS; j++)
	algorithms[current_alg].cleanup(j);
      current_alg = i;
      for (j=0; j<NR_CPUS; j++)
	algorithms[current_alg].init(j);
      break;
    }
    i++;
  }

  return count;
}
                                                                                                                          
/************ avg algorithm variables ******************/

/*** avg_n ***/
/* get/set "n" for the algorithm */
static ssize_t show_avg_n(struct cpufreq_policy *policy, char *buf)
{
  return sprintf (buf, "%d\n", avg_n);
}

static ssize_t
store_avg_n(struct cpufreq_policy *policy, const char *buf, size_t count)
{
  int new_avg_n, ret;

  ret = sscanf(buf, "%d", &new_avg_n);
  if (ret != 1)
    return -EINVAL;
  
  if (new_avg_n >= 0) avg_n = new_avg_n;

  return count;
}

/*** getcurutil_policy ***/
/* get/set 1 of 3 implementations to get utilization info
 */
static ssize_t show_getcurutil(struct cpufreq_policy *policy, char *buf)
{
	if (fp_get_cur_util == get_current_utilization_real) {
		return sprintf (buf, "real_util\n");
	} else if (fp_get_cur_util == get_current_utilization_hype) {
		return sprintf (buf, "hype_util\n");
	} else if (fp_get_cur_util == get_current_utilization_futurehype) {
		return sprintf (buf, "futurehype_util\n");
	} else {
		return sprintf (buf, "error\n");
	}		
}

static ssize_t
store_getcurutil(struct cpufreq_policy *policy, const char *buf, size_t count)
{
  char input[ALG_NAME_LENGTH];
  int ret;

  ret = sscanf(buf, "%s", input);
  if (ret != 1)
    return -EINVAL;
  
  if (input[0] == 'r') {
	  fp_get_cur_util = get_current_utilization_real;
  } else if (input[0] == 'h') {
	  fp_get_cur_util = get_current_utilization_hype;
  } else if (input[0] == 'f') {
	  fp_get_cur_util = get_current_utilization_futurehype;
  }	  

  return count;
}


/************ avg, past, future algorithm variables ******************/

/*** hyst_min ***/
/* get/set for the minimum boundary value 
 * if utilization is below this value we decrease frequency 
 */
static ssize_t show_hyst_min(struct cpufreq_policy *policy, char *buf)
{
  return sprintf(buf, "%d\n", hyst_min);
}

static ssize_t
store_hyst_min(struct cpufreq_policy *policy, const char *buf, size_t count)
{
  int new_hyst_min, ret;

  ret = sscanf(buf, "%d", &new_hyst_min);
  if (ret != 1)
    return -EINVAL;
  
  if (new_hyst_min >= 0 && new_hyst_min < 100 &&
      new_hyst_min < hyst_max) hyst_min = new_hyst_min;

  return count;
}
                                                                                                                      
/*** hyst_max ***/
/* get/set for the minimum boundary value 
 * if utilization is below this value we decrease frequency 
 */
static ssize_t show_hyst_max(struct cpufreq_policy *policy, char *buf)
{
  return sprintf (buf, "%d\n", hyst_max);
}

static ssize_t
store_hyst_max(struct cpufreq_policy *policy, const char *buf, size_t count)
{
  int new_hyst_max, ret;

  ret = sscanf(buf, "%d", &new_hyst_max);
  if (ret != 1)
    return -EINVAL;
  
  if (new_hyst_max >= 0 && new_hyst_max < 100 &&
      new_hyst_max > hyst_min) hyst_max = new_hyst_max;

  return count;
}


/************ past/future algorithm variables ******************/

/*** past_incr ***/
/* get/set for the increment size of past/future algorithm 
 * if utilization is above hyst_max, then increase cpu speed by past_incr.
 */
static ssize_t show_past_incr(struct cpufreq_policy *policy, char *buf)
{
  return sprintf (buf, "%d\n", past_incr);
}

static ssize_t
store_past_incr(struct cpufreq_policy *policy, const char *buf, size_t count)
{
  int ret;
  ret = sscanf(buf, "%d", &past_incr);
  if (ret != 1)
    return -EINVAL;

  return count;
}

/*** past_basesub ***/
/* get/set for the size of subtraction base in past/future algorithm 
 * if utilization is below hyst_min, then use past_basesub as the subtraction
 * base. 
 */
static ssize_t show_past_basesub(struct cpufreq_policy *policy, char *buf)
{
  return sprintf (buf, "%d\n", past_basesub);
}

static ssize_t
store_past_basesub(struct cpufreq_policy *policy, const char *buf, size_t count)
{
  int ret;
  ret = sscanf(buf, "%d", &past_basesub);
  if (ret != 1)
    return -EINVAL;

  return count;
}



/*****************************************************************************/
/* This section contains functions to process kernel dvs data                */
/*****************************************************************************/

/* Function that returns the avg given time slice size for task p */
int avg_orig_time_slice(task_t *p) {
        int i;
	int ctr = 0; // Counts number of nonblank entries
	int sum = 0;
	int avg;

	for (i = 0 ; i < TASKLOADSIZE; i++) {
		/* Skip blank entries */
		if (p->task_load_array[i].orig_time_slice == 0) {
			continue;
		}

		ctr++;
		sum += p->task_load_array[i].orig_time_slice;
	}		

	/* Guard against division by zero */
	if (ctr != 0) {
		avg = sum / ctr;
		if (avg % ctr != 0) { avg++; } /* Round up if necessary */
	} else {
		avg = p->orig_time_slice;
	}

	return avg;
}

/***********************************************************
 * FIXME: Real ugly.
 * The following declarations are ripped from kernel/sched.c
 * I need them for the get_idlecycles_future() function.
 * I don't know of a better solution.
 *
 *
 */

/*
 * These are the runqueue data structures:
 */

#define BITMAP_SIZE ((((MAX_PRIO+1+7)/8)+sizeof(long)-1)/sizeof(long))

typedef struct runqueue runqueue_t;

struct prio_array {
	int nr_active;
	unsigned long bitmap[BITMAP_SIZE];
	struct list_head queue[MAX_PRIO];
};

/*
 * This is the main, per-CPU runqueue data structure.
 *
 * Locking rule: those places that want to lock multiple runqueues
 * (such as the load balancing or the thread migration code), lock
 * acquire operations must be ordered by ascending &runqueue.
 */
struct runqueue {
	spinlock_t lock;
	unsigned long nr_running, nr_switches, expired_timestamp,
			nr_uninterruptible;
	task_t *curr, *idle;
	struct mm_struct *prev_mm;
	prio_array_t *active, *expired, arrays[2];
	int prev_cpu_load[NR_CPUS];
#ifdef CONFIG_NUMA
	atomic_t *node_nr_running;
	int prev_node_load[MAX_NUMNODES];
#endif
	task_t *migration_thread;
	struct list_head migration_queue;

	atomic_t nr_iowait;
};



extern struct runqueue per_cpu__runqueues;

#include <linux/percpu.h>
#define this_rq()		(&__get_cpu_var(runqueues))

/*
 *
 *
 * FIXME: Real ugly.
 * The above declarations are ripped from kernel/sched.c
 *******************************************************/

/* Forware function declaration */
int task_load_calc(task_t *p);

/* Provide information about upcoming jobs in the run queue */
/* Returns a value in jiffies. */
int get_idlecycles_future(void) {
	int idx = 0;
	runqueue_t *rq = this_rq();
	prio_array_t *array = rq->active;
	int flag_checked_active = 0;
	unsigned int total_given = 0;
	unsigned int total_idle = 0;

	/**********************************************************************
	 * Account for remainder of current job
	 *
	 */

	float est_load;
	unsigned int est_runtime;
	unsigned int consumed_sofar;

	/* Adjust utilization for current cpu speed
	 * Pseudo-code:
	 * est_load = (task_load_calc(current) / 100.0) / (dvs_speed / 100.0);
	 * Real code below: */
	est_load = (float) task_load_calc(current) / dvs_speed;

	/* Calculate estimated run time for current job at current speed (since
	 * it is currently running at the current speed). */
	est_runtime = est_load * current->orig_time_slice;

	consumed_sofar = current->orig_time_slice - current->time_slice;

	/* Apply only if task is not taking longer than expected */
	if (est_runtime >= consumed_sofar) {
		unsigned int est_tobe_consumed = est_runtime - consumed_sofar;
		/* Normalize future consumption back to a full speed cpu */
		est_tobe_consumed = est_tobe_consumed * (dvs_speed / 100.0);

		/* Only if there is enough time left do we add the remaining to
		 * the idle time variable */
		if (current->time_slice > est_tobe_consumed) {
			total_idle = current->time_slice - est_tobe_consumed;
		}
	}

	total_given = current->time_slice;

	/*
	 *
 	 * Account for remainder of current job
	 *********************************************************************/

	/* Traverse run queue and find a candidate priority list */
	while (1) {
		struct list_head *headlist, *curlist;
		task_t *curtask;


		/* Find highest priority list with a job */
		if (idx == 0) {
			idx = sched_find_first_bit(array->bitmap);
		} else {
			idx = find_next_bit(array->bitmap, MAX_PRIO, idx);
		}

		/* If run queue has no more jobs for our calculation */
		if (idx >= MAX_PRIO) {
			/* Detect if already tried both active and inactive
			 * queues */
			if (flag_checked_active == 1) {
				/* Account for lack of jobs.
				 * Scale idle to interval size.
				 * Pseudo-code:
				 * total_idle = total_idle / (total_given /
				 *                  interval_length_jiffies);
				 * Real code: */
				total_idle = total_idle *
					interval_length_jiffies / total_given;

				/* The following proved to be too conservative
				 * and gave way too much idle time in some 
				 * initial tests. 
				 * Add remaining idle time to total_idle
				 * total_idle += interval_length_jiffies -
				 *               total_given;
				 */
				break;
			}

			/* Have not tried inactive queue yet.
			 * Set flag and try inactive queue */
			flag_checked_active = 1;
			idx = 0;
			array = rq->expired;
			continue;
		}

		/* Go to target priority list */
		headlist = array->queue + idx;
		curlist = headlist->next;
		curtask = list_entry(curlist, task_t, run_list);

		/* Traverse target priority linked-list */
		while (1) {
			int idlepercent, cur_idle;
			int avg_tslice = avg_orig_time_slice(curtask);

			total_given += avg_tslice;

			/* Prevent overflow. Leads to a conservative idle
			 * estimate */
			if (total_given > interval_length_jiffies) { break; }

			/* Accumulate idle jiffies */
			idlepercent = 100 - task_load_calc(curtask);
			cur_idle = (idlepercent / 100.0) * avg_tslice;
			total_idle += cur_idle;

			/* Next job on this priority list.
			 * For some reason, the following line was incorrect:
			 * cur = list_entry(cur->run_list.prev, task_t,
			 *                  run_list);
			 * The following line and a few appropriate mods solved
			 * the problem. */
			curlist = curlist->next;
			if (curlist == headlist) { break; }
		}

		if (total_given >= interval_length_jiffies) { break; }

		/* Continue searching starting at next priority */
		idx++;

		/* If run queue has no more jobs for our calculation */
		if (idx >= MAX_PRIO) {
			/* Detect if already tried both active and inactive
			 * queues */
			if (flag_checked_active == 1) {
				/* Account for lack of jobs.
				 * Scale idle to interval size.
				 * Pseudo-code:
				 * total_idle = total_idle / (total_given /
				 *                  interval_length_jiffies);
				 * Real code: */
				total_idle = total_idle *
					interval_length_jiffies / total_given;

				/* The following proved to be too conservative
				 * and gave way too much idle time in some
				 * initial tests. 
				 * Add remaining idle time to total_idle
				 * total_idle += interval_length_jiffies -
				 *               total_given;
				 */
				break;
			}

			/* Have not tried inactive queue yet.
			 * Set flag and try inactive queue */
			flag_checked_active = 1;
			idx = 0;
			array = rq->expired;
		}

	}

	if (total_idle > interval_length_jiffies) { 
		total_idle = interval_length_jiffies;
	}
	return total_idle;
}

/* Right now interval is in the range of [-1,1].
 * -1 specifies that you want to use the "pseudo-future" algorithm.
 * For now, I am assuming you mean idle jiffies rather than real cycles.
 */
int get_idlecycles(int interval) {
	int qu_index = dvs_qu_index; /* Index into dvs_qu_array[] */
	int total_jiffies = 0;
	unsigned long long total_idlecycles = 0;
	int divisor;
	int rc;
	int jobctr = 0;
	int skipctr = 0;


	/* Requested information for a "pseudo-future"-based algorithm */
	if (interval == -1) {
		return get_idlecycles_future();
	}


	/* ************************************************** */
	/* Requested information for a "past"-based algorithm */


	/* Work backwards from current index into dvs_qu_array[] */
	while (1) {
		jobctr++;
		qu_index--;
		/* Handle index wrap-around */
		if (unlikely(qu_index < 0)) { qu_index = DVSBUCKETS - 1; }

		/* Prevent potential infinite loop due to wrap-around on
		 * dvs_qu_array[] */
		if (qu_index == dvs_qu_index) { break; }

		/* If not in target interval to gather idle cycles */
		if (total_jiffies < interval_length_jiffies * interval) {
			skipctr++;
			total_jiffies += 
				dvs_qu_array[qu_index].orig_time_slice;

			/* If crossed into target interval */
			if (unlikely(total_jiffies > interval_length_jiffies *
				     interval)) {
				/* Taking an appropriate portion of idle cycles
				 * from nonspillage. */
				int spillage = total_jiffies -
					(interval_length_jiffies * interval);
				int nonspillage =
					dvs_qu_array[qu_index].orig_time_slice
					- spillage;
				/* Do the follow calc, which involves a 64-bit
				 * division.
				 * total_idlecycles += (nonspillage *
				 * dvs_qu_array[qu_index].ns_left) /
				 * dvs_qu_array[qu_index].orig_time_slice;
				 */
				unsigned long long number = nonspillage *
					dvs_qu_array[qu_index].ns_left;
				do_div(number,
				      dvs_qu_array[qu_index].orig_time_slice);
				total_idlecycles += number;

			}

			continue;
		}
		
		/*
		 * From this part onwards, we are in the target interval to
		 * gather idle cycles
		 */

		/* Handle when near interval boundary */
		if (unlikely(total_jiffies +
			     dvs_qu_array[qu_index].orig_time_slice >
			     interval_length_jiffies * (interval + 1))) {
			/* Taking an appropriate portion of idle cycles
			 * from boundary element */
			int remainder = (interval_length_jiffies *
					 (interval + 1)) - total_jiffies;
			if (unlikely(remainder <= 0)) { break; }
			/* Do the follow calc, which involves a 64-bit
			 * division.
			 * total_idlecycles += (remainder *
			 * dvs_qu_array[qu_index].ns_left) /
			 * dvs_qu_array[qu_index].orig_time_slice;
			 */
			unsigned long long number = remainder *
				dvs_qu_array[qu_index].ns_left;
			do_div(number, dvs_qu_array[qu_index].orig_time_slice);
			total_idlecycles += number; 
			break;
		} else {
			/* Handle when not near interval boundary */
			total_jiffies +=
				dvs_qu_array[qu_index].orig_time_slice;
			total_idlecycles += dvs_qu_array[qu_index].ns_left;

			/* If at interval boundary */
			if (unlikely(total_jiffies >=
				     interval_length_jiffies *
				     (interval + 1))) {
				break;
			}
		}
	}

	/* Replaced the following with 3 code lines below to allow for 64-bit
	 * division.
	 * rc = NS_TO_JIFFIES(total_idlecycles);
	 */
	divisor = 1000000000 / HZ;
	do_div(total_idlecycles, divisor);
	rc = total_idlecycles;

	if (unlikely(rc > interval_length_jiffies)) {
		rc = interval_length_jiffies;
	}
	return rc;
}

/* Right now interval is in the range of [-1,1].
 * -1 specifies that you want to use the "pseudo-future" algorithm.
 * For now, I am assuming you mean jiffies rather than real cycles.
 */
int get_runcycles(int interval) {
	return interval_length_jiffies - get_idlecycles(interval);
}

/* Estimated excess cycles based on load estimator.
 * Right now interval is in the range of [-1,1].
 * -1 specifies that you want to use the "pseudo-future" algorithm.
 * For now, I am assuming you mean jiffies rather than real cycles.
 */
int get_excesscycles(int interval) {
	/* I don't know. Return 0 for now. */
	return 0;
}

/* Function that sums up the N load records from the
 * task_load_array and returns the average (an integer between 0 and 100).
 * NOTE: The average is for a full speed cpu.
 * N == TASKLOADSIZE */
int task_load_calc(task_t *p) {
        int i=0;
	int ctr = 0; // Counts number of nonblank entries
        int avg_load = 0;  // compute avg_load. Integer between 0 and 100.
        
	/* Compute the load for ith bin */
        for (i=0; i<TASKLOADSIZE; i++){
		int divisor = 1000000000 / HZ;
		unsigned long long int tmp;
		unsigned int jiffies_left;
		unsigned int run_jiffies;
		float util;
		int util_int;

		/* Skip blank entries */
		if (p->task_load_array[i].orig_time_slice == 0) {
			continue;
		}

		ctr++;

		/* convert ns_left to jiffies */
		tmp = p->task_load_array[i].ns_left;
		do_div(tmp,divisor);
		jiffies_left = (unsigned int) tmp;

		/* Skip if consumed only "bonus" jiffies.
		 * Sometimes, tasks are given a priority boost ("bonus"
		 * jiffies) for sleeping alot. */
		if (jiffies_left >= p->task_load_array[i].orig_time_slice) {
			continue;
		}

		/* Calculate run time jiffies */
		run_jiffies = p->task_load_array[i].orig_time_slice -
			jiffies_left;

		/* Calculate utilization. Get a real between 0 and 1.
		 * I need the float cast because result will be 0.xxxxxxx,
		 * which would otherwise get truncated to 0 before being
		 * assigned to util.
		 */
		util = (float) run_jiffies /
			p->task_load_array[i].orig_time_slice;

		/*****************************************************
		 * Pseudo-code for next few steps to normalize utilization to a
		 * full speed cpu.
		 *
		 * // Convert to an integer between 0 and 100
		 * util_int = util * 100;
		 *
		 * // Convert integer speed to a real between 0 and 1
		 * speed = p->task_load_array[i].speed / 100;
		 *
		 * // Normalize value to a full speed cpu
		 * // Result is an integer between 0 and 100
		 * util_int = util_int * speed;
		 * 
		 * Optimized version in real-code below.
		 * Normalize utilization to a full speed cpu.
		 *****************************************************/
		util_int = util * p->task_load_array[i].speed;

		/* accumulate value */
		avg_load+= util_int;
        }

	/* take the sum of the load and compute average */
	/* Guard against division by zero */
	if (ctr != 0) {
		avg_load = avg_load/ctr;
		/* Round up if necessary */
		if (avg_load % ctr != 0) {
			avg_load++;
		}
	} else {
		avg_load = 0;
	}

        return avg_load;
}



MODULE_AUTHOR("Catalin Drula, David Tam, Winnie Tsang");
MODULE_DESCRIPTION("A dynamic voltage scaling governor");
MODULE_LICENSE("GPL");

module_init(cpufreq_dvs_init);
module_exit(cpufreq_dvs_exit);
